#ifndef _NETLIB_H_INCLUDED_
#define _NETLIB_H_INCLUDED_

/*
 * This file is compatible in FreeBSD-8.3.
 */

#include <stdint.h>
#ifdef USE_MINGW
#if 0
#include <winsock2.h>
#include <ws2tcpip.h>
#endif
#endif

/* Etc */

#define EAFNOSUPPORT EINVAL

typedef unsigned char u_char;
typedef unsigned short u_short;
typedef unsigned int u_int;

/* Socket */

#define AF_UNSPEC 0
#define AF_INET   2
#define AF_INET6 28

typedef uint32_t socklen_t;

struct sockaddr;
int getsockname(int s, struct sockaddr *name, socklen_t *namelen);

/* Link-layer */

#ifdef __linux__
struct sockaddr_ll {
  uint16_t sll_family;
  uint16_t sll_protocol;
  int sll_ifindex;
  uint16_t sll_hatype;
  uint8_t sll_pkttype;
  uint8_t sll_halen;
  uint8_t sll_addr[8];
};
#endif

/* Address */

#define NS_INADDRSZ   4
#define NS_IN6ADDRSZ 16
#define NS_INT16SZ 2

#define INET6_ADDRSTRLEN 46

#define MAXHOSTNAMELEN 256

/* Ethernet */

#define ETHER_ADDR_LEN 6
#define ETHER_TYPE_LEN 2
#define ETHER_HDR_LEN (ETHER_ADDR_LEN * 2 + ETHER_TYPE_LEN)

#define ETHERTYPE_ARP  0x0806
#define ETHERTYPE_IP   0x0800
#define ETHERTYPE_IPV6 0x86DD

struct ether_addr {
  uint8_t octet[ETHER_ADDR_LEN];
} __attribute__((packed));

struct ether_header {
  uint8_t  ether_dhost[ETHER_ADDR_LEN];
  uint8_t  ether_shost[ETHER_ADDR_LEN];
  uint16_t ether_type;
} __attribute__((packed));

/* ARP */

#define ARPHRD_ETHER  1
#ifdef __linux__
#define ARPHRD_LOOPBACK  772
#define ARPHRD_EETHER      2
#define ARPHRD_AX25        3
#define ARPHRD_PRONET      4
#define ARPHRD_CHAOS       5
#define ARPHRD_IEEE802     6
#define ARPHRD_ARCNET      7
#define ARPHRD_FDDI      774
#define ARPHRD_ATM        19
#define ARPHRD_IEEE80211 801
#define ARPHRD_IEEE80211_PRISM    802
#define ARPHRD_IEEE80211_RADIOTAP 803
#define ARPHRD_SLIP      256
#define ARPHRD_PPP       512
#define ARPHRD_TUNNEL    768
#define ARPHRD_NONE   0xFFFE
#endif

#define ARPOP_REQUEST 1
#define ARPOP_REPLY   2

struct arphdr {
  uint16_t ar_hrd;
  uint16_t ar_pro;
  uint8_t  ar_hln;
  uint8_t  ar_pln;
  uint16_t ar_op;
};

/* IP */

#define IPPROTO_ICMP  1
#define IPPROTO_IGMP  2
#define IPPROTO_TCP   6
#define IPPROTO_UDP  17

typedef uint32_t in_addr_t;

struct in_addr {
  in_addr_t s_addr;
};

#define IP_RF (1<<15)
#define IP_DF (1<<14)
#define IP_MF (1<<13)
#define IP_OFFMASK 0x1FFF

struct ip {
#if BYTE_ORDER == LITTLE_ENDIAN
  uint8_t  ip_hl:4, ip_v:4;
#elif BYTE_ORDER == BIG_ENDIAN
  uint8_t  ip_v:4, ip_hl:4;
#else
#error Unknown endian.
#endif
  uint8_t  ip_tos;
  uint16_t ip_len;
  uint16_t ip_id;
  uint16_t ip_off;
  uint8_t  ip_ttl;
  uint8_t  ip_p;
  uint16_t ip_sum;
  struct in_addr ip_src;
  struct in_addr ip_dst;
} __attribute__((packed));

/* IPv6 */

#define IPPROTO_HOPOPTS   0
#define IPPROTO_ROUTING  43
#define IPPROTO_FRAGMENT 44
#define IPPROTO_ICMPV6   58
#define IPPROTO_NONE     59
#define IPPROTO_DSTOPTS  60

struct in6_addr {
  uint8_t s6_addr[16];
};

struct ip6_hdr {
  union {
    uint8_t  ip6_top_vfc;
    uint32_t ip6_top_flow;
  } top;
  uint16_t ip6_plen;
  uint8_t  ip6_nxt;
  uint8_t  ip6_hlim;
  struct in6_addr ip6_src;
  struct in6_addr ip6_dst;
} __attribute__((packed));

#define ip6_vfc  top.ip6_top_vfc
#define ip6_flow top.ip6_top_flow

struct ip6_ext {
  uint8_t ip6e_nxt;
  uint8_t ip6e_len;
} __attribute__((packed));

struct ip6_rthdr {
  uint8_t ip6r_nxt;
  uint8_t ip6r_len;
  uint8_t ip6r_type;
  uint8_t ip6r_segleft;
} __attribute__((packed));

struct ip6_frag {
  uint8_t  ip6f_nxt;
  uint8_t  ip6f_reserved;
  uint16_t ip6f_offlg;
  uint32_t ip6f_ident;
} __attribute__((packed));

#define IP6F_OFF_MASK      ((int)htons(0xFFF8))
#define IP6F_RESERVED_MASK ((int)htons(0x0006))
#define IP6F_MORE_FRAG     ((int)htons(0x0001))

/* ICMP */

#define ICMP_ECHO      8
#define ICMP_ECHOREPLY 0

struct icmp {
  uint8_t  icmp_type;
  uint8_t  icmp_code;
  uint16_t icmp_cksum;
  uint32_t h[1];
  uint32_t n[5];
};

/* ICMPv6 */

#define ICMP6_ECHO_REQUEST 128
#define ICMP6_ECHO_REPLY   129

struct icmp6_hdr {
  uint8_t  icmp6_type;
  uint8_t  icmp6_code;
  uint16_t icmp6_cksum;
  uint32_t icmp6_data32[1];
} __attribute__((packed));

/* IGMP */

struct igmp {
  uint8_t  igmp_type;
  uint8_t  igmp_code;
  uint16_t igmp_cksum;
  struct in_addr igmp_group;
};

/* TCP */

#define TH_FIN  (1<<0)
#define TH_SYN  (1<<1)
#define TH_RST  (1<<2)
#define TH_PUSH (1<<3)
#define TH_ACK  (1<<4)
#define TH_URG  (1<<5)
#define TH_ECE  (1<<6)
#define TH_CWR  (1<<7)

struct tcphdr {
  uint16_t th_sport;
  uint16_t th_dport;
  uint32_t th_seq;
  uint32_t th_ack;
#if BYTE_ORDER == LITTLE_ENDIAN
  uint8_t  th_x2:4, th_off:4;
#elif BYTE_ORDER == BIG_ENDIAN
  uint8_t  th_off:4, th_x2:4;
#else
#error Unknown endian.
#endif
  uint8_t  th_flags;
  uint16_t th_win;
  uint16_t th_sum;
  uint16_t th_urp;
};

/* UDP */

struct udphdr {
  uint16_t uh_sport;
  uint16_t uh_dport;
  uint16_t uh_ulen;
  uint16_t uh_sum;
};

/* Services */

#ifdef USE_MINGW
void bcopy(const void *src, void *dst, unsigned int len);
int bcmp(const void *b1, const void *b2, unsigned int len);
struct timeval;
int select(int fd, void *rfd, void *wfd, void *efd, struct timeval *timeout);
#endif

#if defined(USE_MINGW) || defined(__linux__)
unsigned int strlcpy(char *dst, const char *src, unsigned int size);
#endif

uint16_t htons(uint16_t h);
uint16_t ntohs(uint16_t n);
uint32_t htonl(uint32_t h);
uint32_t ntohl(uint32_t n);

char *ether_ntoa(const struct ether_addr *n);
struct ether_addr *ether_aton(const char *a);
char *inet_ntoa(struct in_addr in);
const char *inet_ntop(int af, const void *src, char *dst, socklen_t size);
int inet_pton(int af, const char *src, void *dst);

#endif
